@@ -28,8 +28,6 @@ class Mock(object):
     def __getattr__(cls, name):
         if name in ('__file__', '__path__'):
             return '/dev/null'
-        elif name[0] == name[0].upper():
-            return type(name, (), {})
         else:
             return Mock()
 # pylint: enable=R0903
@@ -47,6 +45,7 @@ MOCK_MODULES = [
     'cheroot',
     'cheroot.wsgi',
     'cherrypy',
+    'cherrypy.lib',
     'cherrypy.wsgiserver',
     'cherrypy.wsgiserver.ssl_builtin',
 
@@ -61,6 +60,7 @@ MOCK_MODULES = [
     'salt.output',
     'salt.runner',
     'salt.utils',
+    'salt.utils.event',
     'salt.wheel',
 ]
 
