@@ -10,8 +10,29 @@ A REST API for Salt
     authenticating as has permissions for all the functions you will be
     running.
 
-    The configuration options for this module resides in the Salt master config
-    file. All available options are detailed below.
+    Example production configuration block; add to the Salt master config file:
+
+    .. code-block:: yaml
+
+        rest_cherrypy:
+          port: 8000
+          ssl_crt: /etc/pki/tls/certs/localhost.crt
+          ssl_key: /etc/pki/tls/certs/localhost.key
+
+    The REST interface strongly recommends a secure HTTPS connection since Salt
+    authentication credentials will be sent over the wire. If you don't already
+    have a certificate and don't wish to buy one, you can generate a
+    self-signed certificate using the
+    :py:func:`~salt.modules.tls.create_self_signed_cert` function in Salt (note
+    the dependencies for this module):
+
+    .. code-block:: bash
+
+        % salt-call tls.create_self_signed_cert
+
+    All available configuration options are detailed below. These settings
+    configure the CherryPy HTTP server and do not apply when using an external
+    server such as Apache or Nginx.
 
     port
         **Required**
@@ -33,6 +54,26 @@ A REST API for Salt
         will be sent in the clear!
 
         .. versionadded:: 0.8.3
+    thread_pool : ``100``
+        The number of worker threads to start up in the pool.
+
+        .. versionchanged:: 0.8.4
+            Previous versions defaulted to a pool of ``10``
+    socket_queue_size : ``30``
+        Specify the maximum number of HTTP connections to queue.
+
+        .. versionchanged:: 0.8.4
+            Previous versions defaulted to ``5`` connections.
+    max_request_body_size : ``1048576``
+        .. versionchanged:: 0.8.4
+            Previous versions defaulted to ``104857600`` for the size of the
+            request body
+    collect_stats : False
+        Collect and report statistics about the CherryPy server
+
+        .. versionadded:: 0.8.4
+
+        Reports are available via the :py:class:`Stats` URL.
     static
         A filesystem path to static HTML/JavaScript/CSS/image assets.
     static_path : ``/static``
@@ -53,26 +94,11 @@ A REST API for Salt
         useful for apps that utilize the HTML5 history API.
 
         .. versionadded:: 0.8.2
+    root_prefix : ``/``
+        A URL path to the main entry point for the application. This is useful
+        for serving multiple applications from the same URL.
 
-    Example production configuration block:
-
-    .. code-block:: yaml
-
-        rest_cherrypy:
-          port: 8000
-          ssl_crt: /etc/pki/tls/certs/localhost.crt
-          ssl_key: /etc/pki/tls/certs/localhost.key
-
-    The REST interface strongly recommends a secure HTTPS connection since Salt
-    authentication credentials will be sent over the wire. If you don't already
-    have a certificate and don't wish to buy one, you can generate a
-    self-signed certificate using the
-    :py:func:`~salt.modules.tls.create_self_signed_cert` function in Salt (note
-    the dependencies for this module):
-
-    .. code-block:: bash
-
-        % salt-call tls.create_self_signed_cert
+        .. versionadded:: 0.8.4
 
 Authentication
 --------------
@@ -158,19 +184,21 @@ functionality.
 # pylint: disable=W0212,E1101,C0103,R0201,W0221,W0613
 
 # Import Python libs
+import collections
+import itertools
 import functools
 import logging
-import os
 import json
-import textwrap
 
 # Import third-party libs
 import cherrypy
+from cherrypy.lib import cpstats
 import yaml
 
 # Import Salt libs
 import salt
 import salt.auth
+import salt.utils.event
 
 # Import salt-api libs
 import saltapi
@@ -189,7 +217,6 @@ def salt_token_tool():
     if x_auth:
         cherrypy.request.cookie['session_id'] = x_auth
 
-
 def salt_ip_verify_tool():
     '''
     If there is a list of restricted IPs, verify current
@@ -228,7 +255,6 @@ def salt_auth_tool():
     # Session is authenticated; inform caches
     cherrypy.response.headers['Cache-Control'] = 'private'
 
-
 # Be conservative in what you send
 # Maps Content-Type to serialization functions; this is a tuple of tuples to
 # preserve order of preference.
@@ -324,14 +350,7 @@ def urlencoded_processor(entity):
     '''
     # First call out to CherryPy's default processor
     cherrypy._cpreqbody.process_urlencoded(entity)
-    lowdata = entity.params
-
-    # Make the 'arg' param a list if not already
-    if 'arg' in lowdata and not isinstance(lowdata['arg'], list):
-        lowdata['arg'] = [lowdata['arg']]
-
-    # Finally, make a Low State and put it in request
-    cherrypy.request.lowstate = [lowdata]
+    cherrypy.serving.request.unserialized_data = entity.params
 
 
 @process_request_body
@@ -343,7 +362,7 @@ def json_processor(entity):
     '''
     body = entity.fp.read()
     try:
-        cherrypy.serving.request.lowstate = json.loads(body)
+        cherrypy.serving.request.unserialized_data = json.loads(body)
     except ValueError:
         raise cherrypy.HTTPError(400, 'Invalid JSON document')
 
@@ -357,11 +376,28 @@ def yaml_processor(entity):
     '''
     body = entity.fp.read()
     try:
-        cherrypy.serving.request.lowstate = yaml.load(body)
+        cherrypy.serving.request.unserialized_data = yaml.safe_load(body)
     except ValueError:
         raise cherrypy.HTTPError(400, 'Invalid YAML document')
 
 
+@process_request_body
+def text_processor(entity):
+    '''
+    Attempt to unserialize plain text as JSON
+
+    Some large services still send JSON with a text/plain Content-Type. Those
+    services are bad and should feel bad.
+
+    :param entity: raw POST data
+    '''
+    body = entity.fp.read()
+    try:
+        cherrypy.serving.request.unserialized_data = json.loads(body)
+    except ValueError:
+        cherrypy.serving.request.unserialized_data = body
+
+
 def hypermedia_in():
     '''
     Unserialize POST/PUT data of a specified Content-Type.
@@ -378,6 +414,7 @@ def hypermedia_in():
         'application/json': json_processor,
         'application/x-yaml': yaml_processor,
         'text/yaml': yaml_processor,
+        'text/plain': text_processor,
     }
 
     # Do not process the body for POST requests that have specified no content
@@ -392,6 +429,47 @@ def hypermedia_in():
     cherrypy.request.body.processors = ct_in_map
 
 
+def lowdata_fmt():
+    '''
+    Validate and format lowdata from incoming unserialized request data
+
+    This tool requires that the hypermedia_in tool has already been run.
+    '''
+    if cherrypy.request.method.upper() != 'POST':
+        return
+
+    # TODO: call lowdata validation routines from here
+
+    data = cherrypy.request.unserialized_data
+
+    if cherrypy.request.headers['Content-Type'] == 'application/x-www-form-urlencoded':
+        # Make the 'arg' param a list if not already
+        if 'arg' in data and not isinstance(data['arg'], list):
+            data['arg'] = [data['arg']]
+
+        # Finally, make a Low State and put it in request
+        cherrypy.request.lowstate = [data]
+    else:
+        cherrypy.serving.request.lowstate = data
+
+
+cherrypy.tools.salt_token = cherrypy.Tool('on_start_resource',
+        salt_token_tool, priority=55)
+cherrypy.tools.salt_auth = cherrypy.Tool('before_request_body',
+        salt_auth_tool, priority=60)
+cherrypy.tools.hypermedia_in = cherrypy.Tool('before_request_body',
+        hypermedia_in)
+cherrypy.tools.lowdata_fmt = cherrypy.Tool('before_handler',
+        lowdata_fmt, priority=40)
+cherrypy.tools.hypermedia_out = cherrypy.Tool('before_handler',
+        hypermedia_out)
+cherrypy.tools.salt_ip_verify = cherrypy.Tool('before_handler',
+        salt_ip_verify_tool)
+
+
+###############################################################################
+
+
 class LowDataAdapter(object):
     '''
     The primary entry point to the REST API. All functionality is available
@@ -406,11 +484,9 @@ class LowDataAdapter(object):
 
         # 'tools.autovary.on': True,
 
-        'tools.salt_token.on': True,
-        'tools.salt_auth.on': True,
-
         'tools.hypermedia_out.on': True,
         'tools.hypermedia_in.on': True,
+        'tools.lowdata_fmt.on': True,
         'tools.salt_ip_verify.on': True,
     }
 
@@ -418,19 +494,40 @@ class LowDataAdapter(object):
         self.opts = cherrypy.config['saltopts']
         self.api = saltapi.APIClient(self.opts)
 
-    def exec_lowstate(self):
+    def exec_lowstate(self, client=None, token=None):
         '''
         Pull a Low State data structure from request and execute the low-data
         chunks through Salt. The low-data chunks will be updated to include the
         authorization token for the current session.
         '''
         lowstate = cherrypy.request.lowstate
-        token = cherrypy.session.get('token', None)
 
+        # if the lowstate loaded isn't a list, lets notify the client
+        if type(lowstate) != list:
+            raise cherrypy.HTTPError(400, 'Lowstates must be a list')
+
+        # Make any requested additions or modifications to each lowstate, then
+        # execute each one and yield the result.
         for chunk in lowstate:
             if token:
                 chunk['token'] = token
-            yield self.api.run(chunk)
+
+            if client:
+                chunk['client'] = client
+
+            # Make any 'arg' params a list if not already.
+            # This is largely to fix a deficiency in the urlencoded format.
+            if 'arg' in chunk and not isinstance(chunk['arg'], list):
+                chunk['arg'] = [chunk['arg']]
+
+            ret = self.api.run(chunk)
+
+            # Sometimes Salt gives us a return and sometimes an iterator
+            if isinstance(ret, collections.Iterator):
+                for i in ret:
+                    yield i
+            else:
+                yield ret
 
     def GET(self):
         '''
@@ -466,15 +563,13 @@ class LowDataAdapter(object):
             predicate=inspect.ismethod) if not name.startswith('__')]
         clients.remove('run') # run method calls client interfaces
 
-        # Grab a list of output formats
-        formats = [ctype for ctype, _ in ct_out_map]
-
         return {
             'return': "Welcome",
             'clients': clients,
-            'output': formats,
         }
 
+    @cherrypy.tools.salt_token()
+    @cherrypy.tools.salt_auth()
     def POST(self, **kwargs):
         '''
         The primary execution interface for the rest of the API
@@ -529,11 +624,17 @@ class LowDataAdapter(object):
         :status 406: requested Content-Type not available
         '''
         return {
-            'return': list(self.exec_lowstate()),
+            'return': list(self.exec_lowstate(
+                token=cherrypy.session.get('token')))
         }
 
 
 class Minions(LowDataAdapter):
+    _cp_config = dict(LowDataAdapter._cp_config, **{
+        'tools.salt_token.on': True,
+        'tools.salt_auth.on': True,
+    })
+
     def GET(self, mid=None):
         '''
         A convenience URL for getting lists of minions or getting minion
@@ -576,7 +677,8 @@ class Minions(LowDataAdapter):
             'client': 'local', 'tgt': mid or '*', 'fun': 'grains.items',
         }]
         return {
-            'return': list(self.exec_lowstate()),
+            'return': list(self.exec_lowstate(
+                token=cherrypy.session.get('token'))),
         }
 
     def POST(self, **kwargs):
@@ -614,9 +716,6 @@ class Minions(LowDataAdapter):
                 Content-Length: 86
                 Content-Type: application/x-yaml
 
-                - return:
-                    jid: '20130118105423694155'
-
                 return:
                 - jid: '20130603122505459265'
                   minions: [ms-4, ms-3, ms-2, ms-1, ms-0]
@@ -635,21 +734,25 @@ class Minions(LowDataAdapter):
         :status 401: authentication required
         :status 406: requested :mailheader:`Content-Type` not available
         '''
-        for chunk in cherrypy.request.lowstate:
-            chunk['client'] = 'local_async'
-        job_data = list(self.exec_lowstate())
+        job_data = list(self.exec_lowstate(client='local_async',
+            token=cherrypy.session.get('token')))
 
         cherrypy.response.status = 202
         return {
             'return': job_data,
             '_links': {
                 'jobs': [{'href': '/jobs/{0}'.format(i['jid'])}
-                    for i in job_data],
+                    for i in job_data if i],
             },
         }
 
 
 class Jobs(LowDataAdapter):
+    _cp_config = dict(LowDataAdapter._cp_config, **{
+        'tools.salt_token.on': True,
+        'tools.salt_auth.on': True,
+    })
+
     def GET(self, jid=None):
         '''
         A convenience URL for getting lists of previously run jobs or getting
@@ -684,7 +787,7 @@ class Jobs(LowDataAdapter):
                     - '3'
                     Function: test.fib
                     Start Time: 2012, Nov 30 10:46:33.606931
-                    Target: ms-3
+                    Target: jerry
                     Target-type: glob
 
             **Example request**::
@@ -705,27 +808,56 @@ class Jobs(LowDataAdapter):
                 Content-Length: 73
                 Content-Type: application/x-yaml
 
+                info:
+                - Arguments:
+                  - '3'
+                  Function: test.fib
+                  Minions:
+                  - jerry
+                  Start Time: 2012, Nov 30 10:46:33.606931
+                  Target: '*'
+                  Target-type: glob
+                  User: saltdev
+                  jid: '20121130104633606931'
                 return:
-                - ms-3:
+                - jerry:
                   - - 0
                     - 1
                     - 1
                     - 2
-                  - 9.059906005859375e-06
+                  - 6.9141387939453125e-06
 
         :param mid: (optional) a minion id
         :status 200: success
         :status 401: authentication required
         :status 406: requested Content-Type not available
         '''
-        cherrypy.request.lowstate = [{
+        lowstate = [{
             'client': 'runner',
             'fun': 'jobs.lookup_jid' if jid else 'jobs.list_jobs',
             'jid': jid,
         }]
-        return {
-            'return': list(self.exec_lowstate()),
-        }
+
+        if jid:
+            lowstate.append({
+                'client': 'runner',
+                'fun': 'jobs.list_job',
+                'jid': jid,
+            })
+
+        cherrypy.request.lowstate = lowstate
+        job_ret_info = list(self.exec_lowstate(
+            token=cherrypy.session.get('token')))
+
+        ret = {}
+        if jid:
+            job_ret, job_info = job_ret_info
+            ret['info'] = [job_info]
+        else:
+            job_ret = job_ret_info[0]
+
+        ret['return'] = [job_ret]
+        return ret
 
 
 class Login(LowDataAdapter):
@@ -744,16 +876,11 @@ class Login(LowDataAdapter):
     If the request is initiated programmatically, the request must contain a
     :mailheader:`X-Auth-Token` header with valid and active session id.
     '''
-    _cp_config = dict(LowDataAdapter._cp_config, **{
-        'tools.salt_token.on': False,
-        'tools.salt_auth.on': False,
-    })
-
 
     def __init__(self, *args, **kwargs):
         super(Login, self).__init__(*args, **kwargs)
 
-        self.auth = salt.auth.LoadAuth(self.opts)
+        self.auth = salt.auth.Resolver(self.opts)
 
     def GET(self):
         '''
@@ -783,7 +910,6 @@ class Login(LowDataAdapter):
         :status 401: authentication required
         :status 406: requested Content-Type not available
         '''
-        cherrypy.response.status = '401 Unauthorized'
         cherrypy.response.headers['WWW-Authenticate'] = 'Session'
 
         return {
@@ -868,10 +994,14 @@ class Login(LowDataAdapter):
         cherrypy.session['timeout'] = (token['expire'] - token['start']) / 60
 
         # Grab eauth config for the current backend for the current user
-        if token['name'] in self.opts['external_auth'][token['eauth']]:
+        try:
             perms = self.opts['external_auth'][token['eauth']][token['name']]
-        else:
-            perms = self.opts['external_auth'][token['eauth']]['*']
+        except (AttributeError, IndexError):
+            logger.debug("Configuration for external_auth malformed for "\
+                "eauth '{0}', and user '{1}'."
+                .format(token.get('eauth'), token.get('name')), exc_info=True)
+            raise cherrypy.HTTPError(500,
+                'Configuration for external_auth could not be read.')
 
         return {'return': [{
             'token': cherrypy.session.id,
@@ -885,8 +1015,8 @@ class Login(LowDataAdapter):
 
 class Logout(LowDataAdapter):
     _cp_config = dict(LowDataAdapter._cp_config, **{
-        'tools.salt_token.on': False,
-        'tools.salt_auth.on': False,
+        'tools.salt_token.on': True,
+        'tools.salt_auth.on': True,
     })
 
     def POST(self):
@@ -904,19 +1034,8 @@ class Logout(LowDataAdapter):
 class Run(LowDataAdapter):
     _cp_config = dict(LowDataAdapter._cp_config, **{
         'tools.sessions.on': False,
-        'tools.salt_token.on': False,
-        'tools.salt_auth.on': False,
     })
 
-    def exec_lowstate(self):
-        '''
-        Override exec_lowstate to avoid pulling token from the session
-        '''
-        lowstate = cherrypy.request.lowstate
-
-        for chunk in lowstate:
-            yield self.api.run(chunk)
-
     def POST(self, **kwargs):
         '''
         Run commands bypassing the normal session handling
@@ -991,6 +1110,7 @@ class Events(object):
         'tools.encode.encoding': 'utf-8',
 
         # Auth handled manually below
+        'tools.salt_token.on': True,
         'tools.salt_auth.on': False,
 
         'tools.hypermedia_in.on': False,
@@ -1008,11 +1128,18 @@ class Events(object):
 
         .. versionadded:: 0.8.3
 
+        Browser clients currently lack Cross-origin resource sharing (CORS)
+        support for the ``EventSource()`` API. Cross-domain requests from a
+        browser may instead pass the :mailheader:`X-Auth-Token` value as an URL
+        parameter::
+
+            % curl -NsS localhost:8000/events/6d1b722e
+
         .. http:get:: /events
 
             **Example request**::
 
-                % curl -sS localhost:8000/events
+                % curl -NsS localhost:8000/events
 
             .. code-block:: http
 
@@ -1033,14 +1160,47 @@ class Events(object):
 
                 data: {'tag': '20130802115730568475', 'data': {'jid': '20130802115730568475', 'return': True, 'retcode': 0, 'success': True, 'cmd': '_return', 'fun': 'test.ping', 'id': 'ms-1'}}
 
-        .. note:: Caveat when using CORS
+        The event stream can be easily consumed via JavaScript:
 
-            Browser clients currently lack Cross-origin resource sharing (CORS)
-            support for the ``EventSource()`` API. Cross-domain requests from a
-            browser may instead pass the :mailheader:`X-Auth-Token` value as an
-            URL parameter::
+        .. code-block:: javascript
 
-                % curl -sS localhost:8000/events/6d1b722e
+            # Note, you must be authenticated!
+            var source = new EventSource('/events');
+            source.onopen = function() { console.debug('opening') };
+            source.onerror = function(e) { console.debug('error!', e) };
+            source.onmessage = function(e) { console.debug(e.data) };
+
+        It is also possible to consume the stream via the shell.
+
+        Records are separated by blank lines; the ``data:`` and ``tag:``
+        prefixes will need to be removed manually before attempting to
+        unserialize the JSON.
+
+        curl's ``-N`` flag turns off input buffering which is required to
+        process the stream incrementally.
+
+        Here is a basic example of printing each event as it comes in:
+
+        .. code-block:: bash
+
+            % curl -NsS localhost:8000/events |\\
+                    while IFS= read -r line ; do
+                        echo $line
+                    done
+
+        Here is an example of using awk to filter events based on tag:
+
+        .. code-block:: bash
+
+            % curl -NsS localhost:8000/events |\\
+                    awk '
+                        BEGIN { RS=""; FS="\\n" }
+                        $1 ~ /^tag: salt\/job\/[0-9]+\/new$/ { print $0 }
+                    '
+            tag: salt/job/20140112010149808995/new
+            data: {"tag": "salt/job/20140112010149808995/new", "data": {"tgt_type": "glob", "jid": "20140112010149808995", "tgt": "jerry", "_stamp": "2014-01-12_01:01:49.809617", "user": "shouse", "arg": [], "fun": "test.ping", "minions": ["jerry"]}}
+            tag: 20140112010149808995
+            data: {"tag": "20140112010149808995", "data": {"fun_args": [], "jid": "20140112010149808995", "return": true, "retcode": 0, "success": true, "cmd": "_return", "_stamp": "2014-01-12_01:01:49.819316", "fun": "test.ping", "id": "jerry"}}
 
         :status 200: success
         :status 401: could not authenticate using provided credentials
@@ -1071,11 +1231,176 @@ class Events(object):
             yield u'retry: {0}\n'.format(400)
 
             while True:
-                yield u'data: {0}\n\n'.format(json.dumps(stream.next()))
+                data = stream.next()
+                yield u'tag: {0}\n'.format(data.get('tag', ''))
+                yield u'data: {0}\n\n'.format(json.dumps(data))
 
         return listen()
 
 
+class Webhook(object):
+    '''
+    A generic web hook entry point that fires an event on Salt's event bus
+
+    External services can POST data to this URL to trigger an event in Salt.
+    For example, Jenkins-CI or Travis-CI, or GitHub web hooks.
+
+    This entry point does not require authentication. The event data is taken
+    from the request body.
+
+    .. note:: Be mindful of security
+
+        Salt's Reactor can run any code. If you write a Reactor SLS that
+        responds to a hook event be sure to validate that the event came from a
+        trusted source and contains valid data! Pass a secret key and use SSL.
+
+        This is a generic interface and securing it is up to you!
+
+    The event tag is prefixed with ``salt/netapi/hook`` and the URL path is
+    appended to the end. For example, a ``POST`` request sent to
+    ``/hook/mycompany/myapp/mydata`` will produce a Salt event with the tag
+    ``salt/netapi/hook/mycompany/myapp/mydata``. See the :ref:`Salt Reactor
+    <reactor>` documentation for how to react to events with various tags.
+
+    The following is an example ``.travis.yml`` file to send notifications to
+    Salt of successful test runs:
+
+    .. code-block:: yaml
+
+        language: python
+        script: python -m unittest tests
+        after_success:
+            - 'curl -sS http://saltapi-url.example.com:8000/hook/travis/build/success -d branch="${TRAVIS_BRANCH}" -d commit="${TRAVIS_COMMIT}"'
+
+    '''
+    exposed = True
+    tag_base = ['salt', 'netapi', 'hook']
+
+    _cp_config = dict(LowDataAdapter._cp_config, **{
+        # Don't do any lowdata processing on the POST data
+        'tools.lowdata_fmt.on': True,
+
+    })
+
+    def __init__(self):
+        self.opts = cherrypy.config['saltopts']
+        self.event = salt.utils.event.SaltEvent('master',
+                self.opts.get('sock_dir', ''))
+
+    def POST(self, *args, **kwargs):
+        '''
+        Fire an event in Salt with a custom event tag and data
+
+        .. versionadded:: 0.8.4
+
+        .. http:post:: /hook
+
+            **Example request**::
+
+                % curl -sS localhost:8000/hook -d foo='Foo!' -d bar='Bar!'
+
+            .. code-block:: http
+
+                POST /hook HTTP/1.1
+                Host: localhost:8000
+                Content-Length: 16
+                Content-Type: application/x-www-form-urlencoded
+
+                foo=Foo&bar=Bar!
+
+            **Example response**:
+
+            .. code-block:: http
+
+                HTTP/1.1 200 OK
+                Content-Length: 14
+                Content-Type: application/json
+
+                {"success": true}
+
+        As a practical example, an internal continuous-integration build
+        server could send an HTTP POST request to the URL
+        ``http://localhost:8000/hook/mycompany/build/success`` which contains
+        the result of a build and the SHA of the version that was built as
+        JSON. That would then produce the following event in Salt that could be
+        used to kick off a deployment via Salt's Reactor::
+
+            Event fired at Fri Feb 14 17:40:11 2014
+            *************************
+            Tag: salt/netapi/hook/mycompany/build/success
+            Data:
+            {'_stamp': '2014-02-14_17:40:11.440996',
+                'headers': {
+                    'X-My-Secret-Key': 'F0fAgoQjIT@W',
+                    'Content-Length': '37',
+                    'Content-Type': 'application/json',
+                    'Host': 'localhost:8000',
+                    'Remote-Addr': '127.0.0.1'},
+                'post': {'revision': 'aa22a3c4b2e7', 'result': True}}
+
+        Salt's Reactor could listen for the event:
+
+        .. code-block:: yaml
+
+            reactor:
+              - 'salt/netapi/hook/mycompany/build/*':
+                - /srv/reactor/react_ci_builds.sls
+
+        And finally deploy the new build:
+
+        .. code-block:: yaml
+
+            {% set secret_key = data.get('headers', {}).get('X-My-Secret-Key') %}
+            {% set build = data.get('post', {}) %}
+
+            {% if secret_key == 'F0fAgoQjIT@W' and build.result == True %}
+            deploy_my_app:
+              cmd.state.sls:
+                - tgt: 'application*'
+                - arg:
+                  - myapp.deploy
+                  - 'pillar={revision: {{ revision }}}'
+            {% endif %}
+
+        :status 200: success
+        :status 406: requested Content-Type not available
+        :status 413: request body is too large
+        '''
+        tag = '/'.join(itertools.chain(self.tag_base, args))
+        data = cherrypy.serving.request.unserialized_data
+        headers = dict(cherrypy.request.headers)
+
+        ret = self.event.fire_event({
+            'post': data,
+            'headers': headers,
+        }, tag)
+        return {'success': ret}
+
+
+class Stats(object):
+    '''
+    Expose statistics on the running CherryPy server
+    '''
+    exposed = True
+
+    _cp_config = dict(LowDataAdapter._cp_config, **{
+        'tools.salt_token.on': True,
+        'tools.salt_auth.on': True,
+    })
+
+    def GET(self):
+        '''
+        Return a dump of statistics collected from the CherryPy server
+
+        :status 200: success
+        :status 406: requested Content-Type not available
+        '''
+        if hasattr(logging, 'statistics'):
+            return cpstats.extrapolate_statistics(logging.statistics)
+
+        return {}
+
+
 class App(object):
     exposed = True
     def GET(self, *args):
@@ -1095,6 +1420,8 @@ class API(object):
         'run': Run,
         'jobs': Jobs,
         'events': Events,
+        'hook': Webhook,
+        'stats': Stats,
     }
 
     def __init__(self):
@@ -1116,6 +1443,9 @@ class API(object):
             'global': {
                 'server.socket_host': self.apiopts.get('host', '0.0.0.0'),
                 'server.socket_port': self.apiopts.get('port', 8000),
+                'server.thread_pool': self.apiopts.get('thread_pool', 100),
+                'server.socket_queue_size': self.apiopts.get('queue_size', 30),
+                'max_request_body_size': self.apiopts.get('max_request_body_size', 1048576),
                 'debug': self.apiopts.get('debug', False),
             },
             '/': {
@@ -1123,6 +1453,8 @@ class API(object):
 
                 'tools.trailing_slash.on': True,
                 'tools.gzip.on': True,
+
+                'tools.cpstats.on': self.apiopts.get('collect_stats', False),
             },
         }
 
@@ -1153,19 +1485,6 @@ def get_app(opts):
     cherrypy.config['apiopts'] = apiopts
 
     root = API() # cherrypy app
-
     cpyopts = root.get_conf() # cherrypy app opts
 
-    # Register salt-specific hooks
-    cherrypy.tools.salt_token = cherrypy.Tool('on_start_resource',
-            salt_token_tool, priority=55)
-    cherrypy.tools.hypermedia_in = cherrypy.Tool('before_request_body',
-            hypermedia_in)
-    cherrypy.tools.salt_auth = cherrypy.Tool('before_request_body',
-            salt_auth_tool, priority=60)
-    cherrypy.tools.hypermedia_out = cherrypy.Tool('before_handler',
-            hypermedia_out)
-    cherrypy.tools.salt_ip_verify = cherrypy.Tool('before_handler',
-            salt_ip_verify_tool)
-
     return root, apiopts, cpyopts
